from pystreaming.video.enc import Encoder
from pystreaming.video.dist import Distributor
from pystreaming.video.req import Requester
from pystreaming.video.dec import Decoder
from pystreaming.video.handlers import display, collate
from pystreaming.video.patterns import Streamer, Worker, Collector
from pystreaming.video.testimages.imageloader import (
    IMAG_S,
    IMAG_M,
    IMAG_L,
    TEST_S,
    TEST_M,
    TEST_L,
    loadimage,
)



import cv2
from pystreaming.listlib.circulardict import CircularOrderedDict

# handler (generator): Any object that yields frames.
def display(handler, BGR=True):
    for frame, idx in handler:
        if not BGR:
            frame = cv2.cvtColor(frame, cv2.COLOR_RGB2BGR)
        cv2.imshow("pystreaming-display", frame)
        if cv2.waitKey(1) & 0xFF == 27:  # esc pressed
            break


def collate(handler, buffer=10):
    initsize = buffer * 1.2
    collate = CircularOrderedDict(int(initsize))
    for frame, idx in handler:
        if idx == 0:  # Restart collate if stream restart occurs
            collate = CircularOrderedDict(initsize)
        collate.insert_end(idx, frame)
        if len(collate) >= buffer:
            most_recent = min(collate.keys())
            yield collate[most_recent], most_recent
            collate.delete(most_recent)



import zmq, time
import multiprocessing as mp
import pystreaming.video.interface as intf
from pystreaming.listlib.circularlist import CircularList, Empty


# Note. It might be better to implement this as a asyncio system?


def dist_ps(shutdown, infd, endpt, rcvhwm, tracks):
    context = zmq.Context()

    collector = context.socket(zmq.PULL)
    collector.setsockopt(zmq.RCVHWM, rcvhwm)
    collector.bind(infd)

    distributor = context.socket(zmq.REP)
    distributor.bind(endpt)

    if tracks is None:  # default track is "none"
        tracks = ["none"]
    queues = {track: CircularList() for track in tracks}

    while not shutdown.is_set():

        time.sleep(0.001)  # 1000 cycles/sec ~> 4-6 streams at once.

        if collector.poll(0):  # returns 0 if no event, something else if there is
            buf, idx = intf.recv(collector, buf=True, flags=zmq.NOBLOCK)
            for fqueue in queues.values():
                fqueue.push((buf, idx))  # add to buf queue

        if distributor.poll(0):  # got frame req
            track = distributor.recv().decode()
            try:
                buf, idx = queues[track].pop()
                intf.send(distributor, idx, buf=buf, flags=zmq.NOBLOCK)
            except (KeyError, Empty):  # no frames available or wrong track selected
                intf.send(distributor, -1, buf=b"nil", flags=zmq.NOBLOCK)



class Distributor:
    rcvhwm = 30

    def __init__(self, endpoint, seed="", tracks=None):
        self.infd = "ipc:///tmp/encout" + seed
        self.endpoint, self.tracks = endpoint, tracks
        self.shutdown = mp.Event()
        self.ps = None
        self.psargs = (self.shutdown, self.infd, self.endpoint, self.rcvhwm, tracks)

    def start(self):
        if self.ps is not None:
            raise RuntimeError("Tried to start a runnning Distributor")
        
        self.ps = mp.Process(target=dist_ps, args=self.psargs)
        self.ps.daemon = True
        self.ps.start()
        print(self)

    def stop(self):
        if self.ps is None:
            raise RuntimeError("Tried to stop a stopped Distributor")

        self.shutdown.set()
        self.ps.join()
        self.ps = None
        self.shutdown.clear()

    def __repr__(self):
        rpr = ""
        rpr += f"-----Distributor-----\n"
        rpr += f"TRACK: \t{self.tracks}\n"
        rpr += f"IN: \t{self.infd}\n"
        rpr += f"OUT: \t{self.endpoint}\n"
        rpr += f"HWM: \t=IN> {self.rcvhwm})::(XX =OUT> "
        return rpr


from turbojpeg import TurboJPEG, TJSAMP_420, TJFLAG_FASTDCT, TJFLAG_FASTUPSAMPLE
import numpy as np
import zmq, time
import multiprocessing as mp
from functools import partial
from itertools import count
from pystreaming.video.testimages.imageloader import loadimage
import pystreaming.video.interface as intf

QUALITY = 50


def enc_ps(shutdown, infd, outfd, rcvhwm, outhwm):
    context = zmq.Context()

    socket = context.socket(zmq.PULL)
    socket.setsockopt(zmq.RCVHWM, rcvhwm)
    socket.connect(infd)

    out = context.socket(zmq.PUSH)
    out.setsockopt(zmq.SNDHWM, outhwm)
    out.connect(outfd)

    poller = zmq.Poller()
    poller.register(socket, zmq.POLLIN)

    encoder = partial(
        TurboJPEG().encode,
        quality=QUALITY,
        jpeg_subsample=TJSAMP_420,
        flags=TJFLAG_FASTDCT,
    )
    while not shutdown.is_set():
        time.sleep(0.01)  # 100 cycles/sec
        if poller.poll(0):
            try:
                arr, idx = intf.recv(socket, arr=True, flags=zmq.NOBLOCK)
                intf.send(out, idx, buf=encoder(arr), flags=zmq.NOBLOCK)
            except zmq.error.Again:
                # ignore send misses to distributor.
                # Should not happen if there is a distributor.
                # Could implement a signal that there is a distributor. Not strictly necessary tho.
                pass


class Encoder:
    tellhwm = 30
    rcvhwm = outhwm = 10

    def __init__(self, context, seed="", procs=2):
        self.infd = "ipc:///tmp/encin" + seed
        self.outfd = "ipc:///tmp/encout" + seed
        self.context, self.procs = context, procs
        self.shutdown = mp.Event()
        self.psargs = (self.shutdown, self.infd, self.outfd, self.rcvhwm, self.outhwm)
        self.workers = []
        self.idx = 0

        self.sender = self.context.socket(zmq.PUSH)
        self.sender.setsockopt(zmq.SNDHWM, self.tellhwm)
        self.sender.bind(self.infd)

    def send(self, frame):
        if self.workers == []:
            raise RuntimeError("Tried to send frame to stopped Encoder")
        try:
            intf.send(self.sender, self.idx, arr=frame, flags=zmq.NOBLOCK)
            self.idx += 1
        except zmq.error.Again:
            raise RuntimeError("Worker processes are not processing frames fast enough")
            # change behavior to:
            # IF working, but slow, print a warning
            # If not responding, only then raise runtime error

    def start(self):
        if self.workers != []:
            raise RuntimeError("Tried to start running Encoder")
        for _ in range(self.procs):
            self.workers.append(mp.Process(target=enc_ps, args=self.psargs))

        for ps in self.workers:
            ps.daemon = True
            ps.start()

        time.sleep(2)  # Allow workers to spin up
        print(self)

    def stop(self):
        """Stop a pool of encoders.

        Raises:
            RuntimeError: Raised when the Encoder is already stopped
        """
        if self.workers == []:
            raise RuntimeError("Tried to stop stopped Encoder")

        self.shutdown.set()
        for ps in self.workers:
            ps.join()
        self.workers = []
        self.shutdown.clear()

    def _testcard(self, cardenum, animated=False):
        testimage = loadimage(cardenum)
        if self.workers == []:
            self.start()
        if animated:
            storage = []
            for ang in range(360):
                storage.append(np.asarray(testimage.rotate(ang)))
        else:
            storage = np.asarray(testimage)
        for i in count():
            print(i)
            self.send(storage[i % 360] if animated else storage)
            time.sleep(1 / 30)

    def __repr__(self):
        rpr = ""
        rpr += f"-----Encoder-----\n"
        rpr += f"PCS: \t{self.procs}\n"
        rpr += f"IN: \t{self.infd}\n"
        rpr += f"OUT: \t{self.outfd}\n"
        rpr += f"HWM: \t({self.tellhwm} =IN> {self.rcvhwm})::({self.outhwm} =OUT> "
        return rpr


from turbojpeg import TurboJPEG, TJFLAG_FASTDCT, TJFLAG_FASTUPSAMPLE
import zmq, time
from functools import partial
import pystreaming.video.interface as intf
import multiprocessing as mp


def dec_ps(shutdown, infd, outfd, rcvmeta, sndbuf, rcvhwm, outhwm):
    context = zmq.Context()

    socket = context.socket(zmq.PULL)
    socket.setsockopt(zmq.RCVHWM, rcvhwm)
    socket.connect(infd)

    out = context.socket(zmq.PUSH)
    out.setsockopt(zmq.SNDHWM, outhwm)
    out.connect(outfd)

    poller = zmq.Poller()
    poller.register(socket, zmq.POLLIN)

    decoder = partial(TurboJPEG().decode, flags=(TJFLAG_FASTDCT + TJFLAG_FASTUPSAMPLE))
    while not shutdown.is_set():
        time.sleep(0.01)  # 100 cycles a sec
        if poller.poll(0):
            try:
                package = intf.recv(socket, buf=True, meta=rcvmeta, flags=zmq.NOBLOCK)
                if rcvmeta:
                    buf, meta, idx = package
                else:
                    buf, idx = package
                    meta = None
                arr = decoder(buf)
                buf = buf if sndbuf else None
                intf.send(out, idx, arr=arr, buf=buf, meta=meta, flags=zmq.NOBLOCK)
            except zmq.error.Again:
                pass  # ignore send misses to out.


class Decoder:
    hearhwm = 30
    rcvhwm = outhwm = 10

    def __init__(self, context, seed="", rcvmeta=False, sndbuf=False, procs=2):
        self.infd = "ipc:///tmp/decin" + seed
        self.outfd = "ipc:///tmp/decout" + seed
        self.context, self.procs = context, procs
        self.rcvmeta, self.sndbuf = rcvmeta, sndbuf
        self.shutdown = mp.Event()
        self.psargs = (
            self.shutdown,
            self.infd,
            self.outfd,
            rcvmeta,
            sndbuf,
            self.rcvhwm,
            self.outhwm,
        )
        self.workers = []

        self.receiver = self.context.socket(zmq.PULL)
        self.receiver.setsockopt(zmq.RCVHWM, self.hearhwm)
        self.receiver.bind(self.outfd)

    def recv(self):
        if self.workers == []:
            raise RuntimeError("Tried to receive frame from stopped Decoder")
        # Add timeout function?
        return intf.recv(self.receiver, arr=True, buf=self.sndbuf, meta=self.rcvmeta)

    def handler(self):
        if self.workers == []:
            raise RuntimeError("Cannot produce stream handler from stopped Decoder")
        while True:
            yield intf.recv(self.receiver, arr=True, buf=self.sndbuf, meta=self.rcvmeta)

    def start(self):
        if self.workers != []:
            raise RuntimeError("Tried to start running Decoder")
        for _ in range(self.procs):
            self.workers.append(mp.Process(target=dec_ps, args=self.psargs))
        for ps in self.workers:
            ps.daemon = True
            ps.start()

        time.sleep(2)  # allow workers to initialize
        print(self)

    def stop(self):
        if self.workers == []:
            raise RuntimeError("Tried to stop stopped Decoder")

        self.shutdown.set()
        for ps in self.workers:
            ps.join()
        self.workers = []
        self.shutdown.clear()

    def __repr__(self):
        rpr = ""
        rpr += f"-----Decoder-----\n"
        rpr += f"PCS: \t{self.procs}\n"
        rpr += f"IN: \t{self.infd}\n"
        rpr += f"OUT: \t{self.outfd}\n"
        rpr += f"HWM: \t=IN> {self.rcvhwm})::({self.outhwm} =OUT> {self.hearhwm})"
        return rpr


from pystreaming import Encoder, Distributor, Decoder, Requester
import pystreaming.video.interface as intf
import pystreaming
import zmq, uuid
from zmq.devices import ProcessDevice


class Streamer:
    def __init__(self, context, endpoint, procs=2):
        seed = uuid.uuid1().hex
        self.encoder = Encoder(context, seed=seed, procs=procs)
        self.distributor = Distributor(endpoint, seed=seed)
        self.started = False

    def start(self):
        if self.started:
            raise RuntimeError("Tried to start a started Streamer")
        self.started = True
        self.encoder.start()
        self.distributor.start()

    def stop(self):
        if not self.started:
            raise RuntimeError("Tried to stop a stopped Streamer")
        self.started = False
        self.encoder.stop()
        self.distributor.stop()

    def send(self, frame):
        if not self.started:
            raise RuntimeError("Start the Streamer before sending frames")
        self.encoder.send(frame)

    def testsignal(self):
        if not self.started:
            self.start()
        self.encoder._testcard(pystreaming.TEST_M)


class Worker:
    outhwm = 30

    def __init__(self, context, source, drain, track="none", reqprocs=3, decprocs=2):
        seed = uuid.uuid1().hex
        self.requester = Requester(source, seed=seed, track=track, procs=reqprocs)
        self.decoder = Decoder(context, seed=seed, sndbuf=True, procs=decprocs)
        self.drain = context.socket(zmq.PUSH)
        self.drain.setsockopt(zmq.SNDHWM, self.outhwm)
        self.drain.connect(drain)

    def run(self, func):
        try:
            self.decoder.start()
            self.requester.start()
            for arr, buf, idx in self.decoder.handler():
                res = func(arr)
                print(res)
                try:
                    intf.send(self.drain, idx, buf=buf, meta=res, flags=zmq.NOBLOCK)
                except zmq.error.Again:
                    pass  # ignore send misses to drain.

        finally:
            self.requester.stop()
            self.decoder.stop()


class Collector:
    rcvhwm = 30
    outhwm = 30
    
    def __init__(self, context, endpoint):
        seed = uuid.uuid1().hex
        # https://het.as.utexas.edu/HET/Software/pyZMQ/api/zmq.devices.html#zmq.devices.ProcessDevice
        self.device = ProcessDevice(zmq.STREAMER, zmq.PULL, zmq.PUSH)
        self.device.setsockopt_in(zmq.RCVHWM, self.rcvhwm)
        self.device.setsockopt_out(zmq.SNDHWM, self.outhwm)
        self.device.bind_in(endpoint)
        self.device.bind_out("ipc:///tmp/decin" + seed)
        
        
        self.decoder = Decoder(context, seed=seed, rcvmeta=True)

    def handler(self):
        self.device.start()
        self.decoder.start()
        return self.decoder.handler()



from threading import Lock
from collections import OrderedDict
from orderedset import OrderedSet


class Empty(Exception):
    """Raised when trying to pop from an empty queue"""

    def __init__(self, calltype):
        super().__init__(f"Tried to pop from an empty {calltype}")


class CircularList:
    """Fixed size list with push-overwrite capabilities."""

    def __init__(self, maxsize=10):
        """Initialize a Circular List with maximum size.

        Args:
            maxsize (int, optional): Maximum number of elements the list
            can store before it begins to overwrite old elements.
            Defaults to 10.
        """
        self._front = 0  # inclusive
        self._back = 0  # exclusive
        self.size = 0

        self.maxsize = maxsize
        self._array = [None] * maxsize

    def push(self, item):
        """Push an element FIFO (queue) style.

        Args:
            item (pyobj): Object to push.
        """
        if self.size == self.maxsize:  # overwrite old elements
            assert self._front == self._back
            self._array[self._back] = item
            self._back = (self._back + 1) % self.maxsize
            self._front = (self._front + 1) % self.maxsize
        else:
            self._array[self._back] = item
            self._back = (self._back + 1) % self.maxsize
            self.size += 1

    def pop(self):
        """Pop an element off FIFO (queue) style.

        Raises:
            Empty: Raised when there are no elements in the queue

        Returns:
            pyobj: The first element in the list.
        """
        if self.size == 0:
            raise Empty(self.__class__.__name__)
        else:
            ret = self._array[self._front]
            self.size -= 1
            self._front = (self._front + 1) % self.maxsize
            return ret

    def __len__(self):
        return self.size

    def __getitem__(self, idx):
        """Gets the nth element of the list.
        Elements at the front of the list (would be popped off) are at index 0.

        Args:
            idx (int): Index desired to be modified.

        Raises:
            IndexError: Raise if the index is out of range

        Returns:
            pyobj: The object desired.
        """
        if idx >= self.size or idx < 0:
            raise IndexError(f"list index {idx} out of range: [0, {self.size})")
        return self._array[(self._front + idx) % self.maxsize]

    def __setitem__(self, idx, new_val):
        """Sets the nth element of the list.
        Elements at the front of the list (would be popped off) are at index 0.

        Args:
            idx (int): Index desired to be modified.
            new_val (pyobj): New object to emplace.

        Raises:
            IndexError: Raise if the index is out of range
        """
        if idx >= self.size or idx < 0:
            raise IndexError(f"list index {idx} out of range: [0, {self.size})")
        self._array[(self._front + idx) % self.maxsize] = new_val

    def __repr__(self):
        return self._array.__repr__() + f" front: {self._front} back: {self._back}"

    def full(self):
        """Check whether the current list size is the maximum size

        Returns:
            boolean: True if full else False
        """
        return self.size == self.maxsize


# Classes below are either unneeded or need to be updated.


class BufferedCList(CircularList):
    """Subclass of CircularList that introduces the concept of a frame buffer"""

    def __init__(self, *args, buffer=5, **kwargs):
        super().__init__(*args, **kwargs)
        if buffer >= self.maxsize:
            raise ValueError(
                f"buffer:{buffer} must be strictly less than maxsize: {self.maxsize}"
            )
        self.threshold = self.maxsize - buffer

    def at_threshold(self):
        return self.size >= self.threshold


class SafeCList(CircularList):
    """A thread-safe wrapper of the base class CircularList"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.access_lock = Lock()

    def push(self, item):
        with self.access_lock:
            super().push(item)

    def pop(self):
        with self.access_lock:
            return super().pop()

    def __getitem__(self, idx):
        with self.access_lock:
            return super().__getitem__(idx)

    def __setitem__(self, idx, new_val):
        with self.access_lock:
            super().__setitem__(idx, new_val)


class CircularOrderedDict:
    def __init__(self, maxsize):
        self.dict = OrderedDict()
        self.maxsize = maxsize

    def pop_front(self):
        return self.dict.popitem(last=False)

    def insert_end(self, key, value):
        # Insert key-value pair. If key exists, it is moved to the back and updated.
        if key in self.dict:
            self.dict.pop(key)
        self.dict[key] = value
        if len(self.dict) > self.maxsize:
            self.dict.popitem(last=False)

    def delete(self, key):
        del self.dict[key]

    def keys(self):
        return self.dict.keys()

    def __len__(self):
        return len(self.dict)

    def __setitem__(self, key, value):
        if key not in self.dict:
            raise KeyError(key) + " use insert_end to add an element"
        self.dict[key] = value

    def __getitem__(self, key):
        return self.dict[key]

    def __repr__(self):
        return self.dict.__repr__()


class BufferedCOD(CircularOrderedDict):
    def __init__(self, *args, buffer=5, **kwargs):
        super().__init__(*args, **kwargs)
        assert 0 <= buffer and buffer < self.maxsize
        self.threshold = self.maxsize - buffer

    def pop_front(self):
        if len(self.dict) > self.threshold:
            return super().pop_front()
        return None


class CircularOrderedSet(OrderedSet):
    def __init__(self, maxsize):
        super().__init__()
        self.maxsize = maxsize

    def push(self, item):
        if item in self:
            self.remove(item)
        self.add(item)
        while len(self) > self.maxsize:
            self.pop(last=False)
